{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE BangPatterns, CPP #-}
module Main (main) where

import Criterion.Main
import GHC.Exts (inline)

import qualified CArray_2 as A2
import qualified CArray_4 as A4
import qualified CArray_8 as A8
import qualified CArray_16 as A16
import qualified CArray_32 as A32
import qualified CArray_64 as A64
import qualified Constr_2 as C2
import qualified Constr_4 as C4
import qualified Constr_8 as C8
import qualified Constr_16 as C16
import qualified Constr_32 as C32
import qualified Constr_64 as C64
import qualified Control.Monad.ST as ST
import qualified Data.Array as A
import qualified Data.Array.ST as ST
import qualified Data.IntMap as IM
import qualified Data.Sequence as S
import qualified GHC.Arr as ST (unsafeThawSTArray)

main :: IO ()
main = do defaultMain

#define GROUP_INDEX(Name, New, Index)                                                 \
             bgroup Name                                                              \
             [ bgroup "index"                                                         \
               [ bench "06" $ whnf (indexing Index (64::Int)) (New (64::Int) (42::Int))         \
               , bench "10" $ whnf (indexing Index (1024::Int)) (New (1024::Int) (42::Int))     \
               , bench "14" $ whnf (indexing Index (16384::Int)) (New (16384::Int) (42::Int))   \
               , bench "18" $ whnf (indexing Index (262144::Int)) (New (262144::Int) (42::Int)) \
               ]                                                                      \
             ]

#define GROUP_CUST(Name, New, Indexing, Updating)                                     \
             bgroup Name                                                              \
             [ bgroup "index"                                                         \
               [ bench "06" $ whnf (Indexing (64::Int)) (New (64::Int) (42::Int))         \
               , bench "10" $ whnf (Indexing (1024::Int)) (New (1024::Int) (42::Int))     \
               , bench "14" $ whnf (Indexing (16384::Int)) (New (16384::Int) (42::Int))   \
               , bench "18" $ whnf (Indexing (262144::Int)) (New (262144::Int) (42::Int)) \
               ]                                                                      \
             , bgroup "update"                                                        \
               [ bench "06" $ whnf (Updating (64::Int)) (New (64::Int) (42::Int))          \
               , bench "10" $ whnf (Updating (1024::Int)) (New (1024::Int) (42::Int))      \
               , bench "14" $ whnf (Updating (16384::Int)) (New (16384::Int) (42::Int))    \
               , bench "18" $ whnf (Updating (262144::Int)) (New (262144::Int) (42::Int))  \
               ]                                                                      \
             ]

#define GROUP(Name, New, Index, Update) GROUP_CUST(Name, New, indexing Index, updating Update)

           [ GROUP("A_2", A2.new, A2.index, A2.update)
           , GROUP("A_4", A4.new, A4.index, A4.update)
           , GROUP("A_8", A8.new, A8.index, A8.update)
           , GROUP("A_16", A16.new, A16.index, A16.update)
           , GROUP("A_32", A32.new, A32.index, A32.update)
           , GROUP("A_64", A64.new, A64.index, A64.update)
           , GROUP("C_2", C2.new, C2.index, C2.update)
           , GROUP("C_4", C4.new, C4.index, C4.update)
           , GROUP("C_8", C8.new, C8.index, C8.update)
           , GROUP("C_16", C16.new, C16.index, C16.update)
           , GROUP("C_32", C32.new, C32.index, C32.update)
           , GROUP("C_64", C64.new, C64.index, C64.update)
           , GROUP_INDEX("Mono", a_new, a_index)
           , GROUP_CUST("STA", st_new, indexingST, updatingST)
           , GROUP("IntMap", im_new, im_index, IM.insert)
           , GROUP("Seq", s_new, s_index, S.update)
           ]

indexing index' !len = \arr -> indexing' arr 0 0
  where indexing' arr !acc !i | i >= len = acc
                              | otherwise = indexing' arr (acc + inline index' i arr) (i+1)
{-# INLINE indexing #-}

updating update' !len = \arr -> updating' 0 arr
  where updating' !i !arr | i >= len = arr
                          | otherwise = updating' (i+1) (inline update' i 43 arr)
{-# INLINE updating #-}

a_new len a = A.listArray (0, len-1) (replicate len a)
{-# INLINE a_new #-}
a_index i arr = arr A.! i
{-# INLINE a_index #-}

im_new len a = IM.fromList (zip [0..] (replicate len a))
{-# INLINE im_new #-}
im_index i im = im IM.! i
{-# INLINE im_index #-}

s_new len a = S.fromList (replicate len a)
{-# INLINE s_new #-}
s_index i s = S.index s i
{-# INLINE s_index #-}

st_new len a = ST.runSTArray (ST.newArray (0, len-1) a)
{-# INLINE st_new #-}
indexingST !len = \arr -> ST.runST (ST.unsafeThawSTArray arr >>= \arr -> indexing' arr 0 0)
  where indexing' arr !acc !i | i >= len = return acc
                              | otherwise = do inc <- ST.readArray arr i
                                               indexing' arr (acc + inc) (i+1)
{-# INLINE indexingST #-}

updatingST !len = \arr -> ST.runSTArray (ST.unsafeThawSTArray arr >>= updating' 0)
  where updating' !i !arr | i >= len = return arr
                          | otherwise = do ST.writeArray arr i 43
                                           updating' (i+1) arr
{-# INLINE updatingST #-}
